theory QImp1 imports Main begin

text{* Implementation for the Queues *}

text{* 
  The queue is a list, with the front of the list as the front of the queue.
  This leads to pleasingly simple proofs, but is rather inefficient.
  *}
 

datatype 'a q = Q "'a list" 

definition mtQ   :: "'a q"
where "mtQ = Q []"

fun isEmpty :: "'a q => bool"
where "isEmpty (Q xs) = List.null xs"

fun enq     :: "'a => 'a q => 'a q"
where "enq x (Q xs) = Q (xs @[x])"

fun front   :: "'a q => 'a"
where "front (Q xs) = hd xs"

fun deq     :: "'a q => 'a q"
where "deq (Q xs) = Q (tl xs)"

text{*
  We can now show the six axioms of the requirement specification.
  *}

lemma empty: "isEmpty mtQ"
  by (simp add: mtQ_def null_def)

lemma notEmpty: "~ (isEmpty (enq a q))"
  apply (induction q)
  by (simp add: mtQ_def null_def)

lemma front_mt: "isEmpty q \<Longrightarrow> front (enq a q) = a"
  apply (induction q)
  apply (simp add: null_def)
  done

lemma front_nonmt: "\<not> isEmpty q \<Longrightarrow> front (enq a q) = front q"
  apply (induction q)
  apply (simp add: null_def)
  done

(* declare [[ simp_trace = true ]] *)

lemma deq_mt: "isEmpty q \<Longrightarrow> deq (enq a q) = mtQ"
  apply (induction q)
  apply (simp add: null_def mtQ_def)
  done

lemma deq_nonmt: "\<not> isEmpty q \<Longrightarrow> deq (enq a q) = enq a (deq q)"
  apply (induction q)
  apply (simp add: null_def)
  done

end
