theory QImp3 imports Main begin

text{* Implementation for the Queues. *}

text{*
  The queue is a list, with the front of the list as the back of the queue.
  It leads to longer proofs, and is inefficient. 
  *}

datatype 'a q = Q "'a list"

definition mtQ :: "'a q"
where "mtQ == Q []"

fun isEmpty :: "'a q=> bool"
where "isEmpty (Q q)= List.null q"

fun enq     :: "'a => 'a q => 'a q"
where "enq x (Q q)= Q (x#q)"

fun front   :: "'a q => 'a"
where "front (Q q)= hd (rev q)"

fun deq     :: "'a q => 'a q"
where "deq (Q q)= Q (rev (tl (rev q)))"

lemma empty: "isEmpty mtQ"
  apply (unfold mtQ_def)
     by (simp add: null_def)

lemma notEmpty: "~ (isEmpty (enq a q))"
  apply (induct "q")
  apply (auto simp add: null_def)
  done

lemma rev_null [simp]: "List.null xs ==> rev xs= []"
  apply (induct "xs")
  apply (auto simp add: null_def)
  done

lemma rev_not_null [simp]: "hd (rev (xs @ [a]))= a"
  apply (induct "xs")
  apply (auto)
  done
 
lemma not_null [simp]: "~ List.null (xs @ [x])"
  apply (cases "xs")
  apply (simp_all add: null_def)
  done

lemma rev_pres_not_null [simp]: "~ List.null xs ==> ~ List.null (rev xs)"
  apply (cases "xs")
  apply (simp_all add: null_def) 
  done

lemma hd_not_null [simp]: "~ List.null xs ==> hd (xs @ ys)= hd xs"
  apply (cases "xs")
  apply (simp_all add: null_def)
  done

lemma front: "front (enq a q) = (if isEmpty q then a else front q)"
  apply (induct "q")
  apply (auto)
  done
  
lemma deq: "deq (enq a q) = (if isEmpty q then mtQ else enq a (deq q))"
  apply (induct "q")
  apply (auto simp add: null_def mtQ_def)
  done

text {* The alternative, shorter definitions

fun front :: "'a q => 'a"
where "front (Q q)= last q"

fun deq   :: "'a q => 'a q"
where "deq (Q q)= Q (butlast q)"

do not work, because there is no lemma that e.g @term{last q = hd (rev q)}.
Could prove that, I suppose.
*}

end


